import torch
import torch.nn as nn
import torch.distributions as distributions
from . import common
from .. import device


class Sampler(nn.Module):
    def __init__(self, backflow = False, mean_scale = 1.0, stddev_scale = 1.0, distrib_type = "normal"):
        self.backflow = backflow
        self.mean_scale = device.tensor([mean_scale])
        self.stddev_scale = device.tensor([stddev_scale])
        if distrib_type not in ["normal", "uniform"]:
            raise Exception("Unknown distribution type %s, should be normal or uniform" % self.distrib_type)

        self.distrib_type = distrib_type
        super(Sampler, self).__init__()
        
    def forward(self, batch):
        size = batch.shape[1] // 2

        if self.distrib_type == "normal":
            distrib = distributions.Normal(batch[:,::2] * self.mean_scale, torch.abs(batch[:,1::2]) * self.stddev_scale)
        elif self.distrib_type == "uniform":
            center = batch[:,::2] * self.mean_scale
            half_size = torch.abs(batch[:,1::2]) * self.stddev_scale * 0.5

            low = center - half_size
            high = center + half_size
            
            distrib = distributions.Uniform(low, high)
            
        if self.backflow:
            return distrib.rsample()
        else:
            return distrib.sample()
    
class SamplerGen(common.LayerGen):
    def __init__(self, backflow = False, mean_scale = 1.0, stddev_scale = 1.0, distrib_type = "normal"):
        self.mean_scale = mean_scale
        self.stddev_scale = stddev_scale
        self.backflow = backflow
        self.distrib_type = distrib_type
        if distrib_type not in ["normal", "uniform"]:
            raise Exception("Unknown distribution type %s, should be normal or uniform" % self.distrib_type)

        super(SamplerGen, self).__init__()
        
    def build(self, input_size):
        layer = Sampler(self.backflow, self.mean_scale, self.stddev_scale, self.distrib_type)
        return layer, (input_size[0] // 2,)

class GaussianDistribution(nn.Module):
    def __init__(self, gaussian_parameters, *args, **kwargs):
        super(GaussianDistribution, self).__init__(*args, **kwargs)
        self.gaussian_parameters = (torch.Tensor(gaussian_parameters[0]), torch.Tensor(gaussian_parameters[1]))
        
    def forward(self, batch):
        ret = distributions.Normal(*self.gaussian_parameters).sample()
        return ret
    
class GaussianDistributionGen(common.LayerGen):
    def __init__(self, gaussian_parameters, *args, **kwargs):
        self.gaussian_parameters = gaussian_parameters
        super(GaussianDistributionGen, self).__init__(*args, **kwargs)
        
    def build(self, input_size):
        layer = GaussianDistribution(self.gaussian_parameters)
        assert(len(self.gaussian_parameters[0]) == len(self.gaussian_parameters[1]))
        return layer, (len(self.gaussian_parameters[0]),)


class UniformDistribution(nn.Module):
    def __init__(self, uniform_parameters, *args, **kwargs):
        super(UniformDistribution, self).__init__(*args, **kwargs)
        self.uniform_parameters = (torch.Tensor(uniform_parameters[0]), torch.Tensor(uniform_parameters[1]))
        
    def forward(self, batch):
        ret = distributions.Uniform(*self.uniform_parameters).sample()
        return ret
    
class UniformDistributionGen(common.LayerGen):
    def __init__(self, uniform_parameters, *args, **kwargs):
        self.uniform_parameters = uniform_parameters
        super(UniformDistributionGen, self).__init__(*args, **kwargs)
        
    def build(self, input_size):
        layer = UniformDistribution(self.uniform_parameters)
        assert(len(self.uniform_parameters[0]) == len(self.uniform_parameters[1]))
        return layer, (len(self.uniform_parameters[0]),)    



import torch

device = None
cuda_enabled = False

def prepare_object(o):
    global device, cuda_enabled
    if device == None and cuda_enabled:
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        print("Running on device", device)
        
    return o.to(device)

def prepare_numpy_object(o):
    return prepare_object(torch.FloatTensor(o))

def zeros(size):
    global device
    return torch.zeros(size, device = device)

def tensor(tensor, *args, **kwargs):
    ret = torch.tensor(tensor, *args, dtype=torch.float32, **kwargs)
    return prepare_object(ret)

def tensor2numpy(t):
    return t.cpu().data.numpy()






from . import common
import torch
import torch.nn as nn
import torchgeometry.contrib
from .. import device

class ReLUGen(common.LayerGen):
    Layer = nn.ReLU

class LeakyReLUGen(common.LayerGen):
    Layer = nn.LeakyReLU

class SigmoidGen(common.LayerGen):
    Layer = nn.Sigmoid

class TanhGen(common.LayerGen):
    Layer = nn.Tanh

class LogSigmoidGen(common.LayerGen):
    Layer = nn.LogSigmoid

class SoftplusGen(common.LayerGen):
    Layer = nn.Softplus

class Square(nn.Module):
    def forward(self, batch):
        return torch.mul(batch, batch)

class SquareGen(common.LayerGen):
    Layer = Square
    
class Abs(nn.Module):
    def forward(self, batch):
        return torch.abs(batch)
    
class AbsGen(common.LayerGen):
    Layer = Abs
    
class Exp(nn.Module):
    def forward(self, batch):
        return torch.exp(batch)

class ExpGen(common.LayerGen):
    Layer = Exp

class Gauss(nn.Module):
    def forward(self, batch):
        return torch.exp(-batch * batch)

class GaussGen(common.LayerGen):
    Layer = Gauss
    
class LogSoftmaxGen(common.LayerGen):
    Layer = nn.LogSoftmax
    
class SoftmaxGen(common.LayerGen):
    Layer = nn.Softmax
    
class Softmax2dGen(common.LayerGen):
    Layer = nn.Softmax2d

class SoftmaxSurface(nn.Module):
    def __init__(self, mode = None, *args, **kwargs):
        self.mode = mode
        super(SoftmaxSurface, self).__init__(*args, **kwargs)
        
    def forward(self, batch):
        shape = [s for s in batch.shape]
        if self.mode == "exp2":
            shape[1] *= 2
            
        ret = device.zeros(shape)
        for i in range(batch.shape[0]):
            for j in range(batch.shape[1]):
                e = batch[i][j]
                
                if self.mode in ["exp2", "exp2"]:
                    e = torch.exp(e - torch.max(e))
                    s = torch.sum(e)
                    
                    if s.item() != 0:
                        if self.mode == "exp2":
                            ret[i][2 * j] = e / s
                        else:
                            ret[i][j] += e / 2.0

                    e = torch.exp(- e + torch.min(e))
                    s = torch.sum(e)
                    
                    if s.item() != 0:
                        if self.mode == "exp2":                            
                            ret[i][2 * j + 1] = e / s
                        else:
                            ret[i][j] += e / (s * 2.0)
                elif self.mode in ["square"]:
                    m = torch.mean(e)
                    e = (e - m) ** 2
                    s = torch.sum(e)
                    
                    if s.item() != 0:
                        ret[i][j] = e / s
                        
                else:
                    s = torch.sum(torch.abs(e))
                    
                    if s.item() != 0:
                        ret[i][j] = e / s               
        return ret

class SoftmaxSurfaceGen(common.LayerGen):
    def __init__(self, mode = None, *args, **kwargs):
        self.mode = mode
        super(SoftmaxSurfaceGen, self).__init__(*args, **kwargs)
        
    def build(self, input_size):
        layer = SoftmaxSurface(self.mode, *self.args, **self.kwargs)

        output_size = [s for s in input_size]
        if self.mode == "exp2":
            output_size[0] *= 2
            return layer, output_size
        else:
            return layer, output_size

    
class SpatialSoftArgmax2dGen(common.LayerGen):
    def build(self, input_size, *args, **kwargs):
        layer = torchgeometry.contrib.SpatialSoftArgmax2d(*args, **kwargs)
        return layer, (2,)






import torch
import torch.nn as nn
from . import conv
from .. import device

class MaxPool2dGen(conv.ConvLikeBaseGen):
    # If kernel size has -1 somewhere, it means we should collapse the given dimension
    def __init__(self, kernel_size = None, target_size = None, *args, **kwargs):
        super(MaxPool2dGen, self).__init__(*args, **kwargs)
        self.kernel_size = kernel_size
        self.target_size = target_size

    def build(self, input_size):
        if self.target_size != None:
            self.kernel_size = [s // self.target_size for s in input_size]        

        kernel_size = self.kernel_size_helper(self.kernel_size, input_size)
        layer = nn.MaxPool2d(kernel_size, *self.args, **self.kwargs)

        return self.build_helper(layer, input_size)


class AvgPool2dGen(conv.ConvLikeBaseGen):
    # If kernel size has -1 somewhere, it means we should collapse the given dimension
    def __init__(self, kernel_size = None, *args, target_size = None, **kwargs):
        super(AvgPool2dGen, self).__init__(*args, **kwargs)
        self.kernel_size = kernel_size
        self.target_size = target_size

    def compute_size(self, layer, Hin, Win, kernel_size, padding, stride, dilation):
        Hout = (Hin + 2 * padding[0] - kernel_size[0]) // stride[0] + 1
        Wout = (Win + 2 * padding[1] - kernel_size[1]) // stride[1] + 1
        return Hout, Wout

    def build(self, input_size):
        if self.target_size != None:
            self.kernel_size = [s // self.target_size for s in input_size]
                    
        kernel_size = self.kernel_size_helper(self.kernel_size, input_size)
        layer = nn.AvgPool2d(kernel_size, *self.args, **self.kwargs)

        ret = self.build_helper(layer, input_size)
        return ret

class WinnerTakeAll2D(nn.Module):
    def __init__(self, input_size, avg_kernel_size, *args, **kwargs):
        super(WinnerTakeAll2D, self).__init__(*args, **kwargs)
        self.input_size = input_size
        self.avg_kernel_size = avg_kernel_size
            
        self.max_pool_kernel_size = (input_size[1] // avg_kernel_size[0], input_size[2] // avg_kernel_size[1])
#        print(self.max_pool_kernel_size, input_size, avg_kernel_size)
        
        self.avg_pool = nn.AvgPool2d(kernel_size = self.avg_kernel_size)
        self.max_pool = nn.MaxPool2d(kernel_size = self.max_pool_kernel_size, return_indices = True)
        self.xy_grid = None

    def forward(self, batch):
        output = self.avg_pool(batch)
        maxoutput, maxpos = self.max_pool(torch.abs(output))
        index = maxpos.item()
        
        if self.xy_grid is None:
            x_size = batch.shape[2]
            y_size = batch.shape[3]
            x_cord = torch.arange(x_size)
            x_grid = x_cord.repeat(y_size).view((x_size, y_size))
            y_cord = torch.arange(y_size)
            y_grid = y_cord.repeat(x_size).view((y_size, x_size)).t()
            
            self.xy_grid = torch.stack([x_grid, y_grid], dim=-1).type(torch.FloatTensor)

        if batch.shape[2] != batch.shape[3]:
            assert(False) # This case was not tested, x_index,y_index and variance may have to be changed
            
        x_index = (index % output.shape[2]) * self.avg_kernel_size[0]
        y_index = (index // output.shape[2]) * self.avg_kernel_size[1]

        std_dev = float(batch.shape[2] / self.avg_kernel_size[0]) / 3.0

        mean = device.tensor([x_index, y_index])
        gaussian = -torch.sum((self.xy_grid - mean)**float(2.0), dim=-1)
        gaussian /= (std_dev ** 2) * 2.0
        gaussian = torch.exp(gaussian)
#        gaussian /= torch.sum(gaussian)

        gaussian = gaussian.unsqueeze(0).unsqueeze(0)
        return gaussian
        
class WinnerTakeAll2DGen(conv.ConvLikeBaseGen):
    def __init__(self, target_size = None, kernel_size = None, *args, **kwargs):
        super(WinnerTakeAll2DGen, self).__init__(*args, **kwargs)
        self.kernel_size = kernel_size
        self.target_size = target_size
    
    def build(self, input_size):
        if self.target_size != None:
            if isinstance(self.target_size, int):
                self.target_size = [self.target_size] * 2
            self.kernel_size = []
            for i, s in enumerate(input_size[1:]):
                self.kernel_size += [s // self.target_size[i]]
                    
        kernel_size = self.kernel_size_helper(self.kernel_size, input_size)

        for i, s in enumerate(input_size[1:]):
            if s % kernel_size[i] != 0:
                raise Exception("The input size %s is not a multiple of kernel_size %s" % (input_size, kernel_size))                
        
        net = WinnerTakeAll2D(input_size, avg_kernel_size = (4,4))

#        output_size = (input_size[0], input_size[1] // kernel_size[0], input_size[2] // kernel_size[1])
#        print(input_size, output_size)

        return net, input_size



class LayerGen(object):
    def __init__(self, *args, **kwargs):
        super(LayerGen, self).__init__()
        self.args = args
        if "post_hook" in kwargs:
            self.post_hook = kwargs["post_hook"]
            del kwargs["post_hook"]
        else:
            self.post_hook = None
        self.kwargs = kwargs

    def build(self, input_size):
        layer = self.Layer(*self.args, **self.kwargs)
        return layer, input_size

    def get_input_names(self):
        return None

    def full_name(self):
        if hasattr(self, "parent"):
            name = self.parent.find_child_name(self)
            parent_name = self.parent.full_name()
            if parent_name is not None:                
                return parent_name + "." + name
            else:
                return name
        else:
            return None

    def build_helper(self, layer, input_size):
#        print ("build_helper", input_size)
        Hin = input_size[1]
        Win = input_size[2]
        Cin = input_size[0]
        if not hasattr(layer, "out_channels"):
            Cout = Cin
        else:
            Cout = layer.out_channels

        if hasattr(layer, "dilation"):
            dilation = layer.dilation
        else:
            dilation = 0
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        padding = layer.padding
        if isinstance(padding, int):
            padding = (padding, padding)

        stride = layer.stride
        if isinstance(stride, int):
            stride = (stride, stride)

        Hout, Wout = self.computeSize(layer, Hin, Win, layer.kernel_size, padding, stride, dilation)
        output_size = (Cout, Hout, Wout)

        # TODO : move this to the right place
        if self.post_hook:
            self.post_hook(layer, output_size)

        return layer, output_size

    def compute_size(self, layer, Hin, Win, kernel_size, padding, stride, dilation):
        raise NotImplementedError("%s should implement compute size" % self.__class__.__name__ )



from functools import reduce

def product(elements):
    return reduce(lambda x,y : x * y, elements)

def flat_numpy(tensor):
    return tensor.squeeze().data.numpy()

# Utility class to use python call syntax to create elegant ordered dict
class od(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
    
    def __iter__(self):
        return iter(self.kwargs.items())



import os
import sys
import time
import datetime
import numpy
from . import world, dashboard, config
from matplotlib import pyplot
import os.path
import traceback

class Experiment(object):
    # TODO : override this in subclasses
    trainer_factory = None

    @staticmethod
    def aura_xp_data_path():
        return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "aura_xp_data"))

    def __init__(self, size = 320, base_path = None, config_path = None):
        self.world = world.World(self, (size,size))
        self.debug_images = {}
        self.windows = {}
        self.current_offset = [00, 0]
        if base_path == None:
            base_path = self.aura_xp_data_path()
        self.base_path = base_path

        # load config from config_path, or if it is None, from experiment python file itself, calling fill_config function
        self.config = config.Config(config_path or sys.modules[self.__module__].__file__)
        self.config.set("xp/name", self.get_xp_name())
        self.dashboard = dashboard.Dashboard(self.config, dashboard.MPLBackend(self.config))

        self.dashboard_last_update_time = 0.0

    def create_world(self):
        pass

    def create_organism(self):
        pass

    def update(self, time):
        self.trainer.update(time)

    def post_update(self, time):
        self.trainer.post_update(time)

    def display_disable(self):
        return self.config.get("dashboard/display_disable", False)

    def accept_display(self, timestamp):
        ret = self.dashboard.accept_display(timestamp)
        return ret
    
    def start_training_sample(self, timestamp):
        """Used to signal we are starting a new training sample"""
        try:
            self.config.reload()
        except Exception as e:
            traceback.print_exc()
        max_elapsed = self.config.get("dashboard/refresh", 1.0)

        t = time.time()
        elapsed = t - self.dashboard_last_update_time

        if self.display_disable():
            return
        
        if elapsed >= max_elapsed:
            self.dashboard_last_update_time = t
            self.dashboard.start_sequence(timestamp)
        else:
            self.dashboard.display_refresh()

    def finish_training_sample(self):
        self.dashboard.end_sequence()

    def add_display(self, timestamp, image_name, image):
        try:
            self.dashboard.add_display(timestamp, image_name, image)
        except Exception as e:
            print("Could not display %s" % image_name)
            traceback.print_exc()

    def get_xp_name(self):
        import sys
        f = os.path.split(sys.modules[self.__module__].__file__)[-1]
        xp_name = os.path.splitext(f)[0]
        return xp_name

    def get_xp_module(self):
         # TEMPORARY
        return "fl_ml.aura.xps." + self.get_xp_name()

    def get_latest_path(self):
        return self.get_latest_path_for_name(self.get_xp_name())

    @staticmethod
    def get_latest_path_for_name(xp_name):
        input_dir = os.path.abspath(os.path.join(Experiment.aura_xp_data_path(), "xps", xp_name))

        # Get the last subdirectories (the dir names all finish by a date)
        dates = [os.path.join(input_dir, f) for f in os.listdir(input_dir)]
        dates = [d for d in dates if os.path.isdir(d)]
        dates.sort(reverse = True)

        max_dir = None

        for last in dates:
            max_epoch = -1
            for g in os.listdir(last):
                if not os.path.isdir(os.path.join(last, g)):
                    continue
                epoch = -1
                try:
                    epoch = int(g.split("_")[1])
                except Exception as e:
                    print("Could not parse %s : %s" % (g, e))
                if epoch > max_epoch:
                    max_dir = os.path.join(last, g)
                    max_epoch = epoch
            if max_dir != None:
                return max_dir
            else:
                print("The directory did not contain a valid network, maybe was interrupted early: %s" % last)

    def run(self, max_time):
        print("----------- STARTING -----------")
        directory = [self.get_xp_name()]
        directory += [str(datetime.datetime.now()).replace(" ", "_").replace(":", "-")]

        full_path = os.path.abspath(os.path.join(self.base_path, "xps", *directory))
        self.save_path = full_path
        print("SAVE PATH IS", self.save_path)

        for obj in self.create_world():
            if obj != None:
                self.world.add_child(obj)

        self.organism = self.create_organism()
        self.world.add_child(self.organism)

        self.trainer = self.trainer_factory(self)

        for i in range(0, max_time, 1):
            self.update(i)
            self.world.update(i)
            self.post_update(i)

            if (i % self.config.get("network_save_epochs", 10000)) == 0 or i == (max_time - 1):
                full_path0 = os.path.join(self.save_path, "epoch_" + str(i))
                os.makedirs(full_path0)
                self.organism.save(full_path0)
                self.trainer.save(self.save_path)

                full_path_screenshot = os.path.join(full_path0, "screenshot.png")
                self.dashboard.display_refresh(full_path_screenshot, self.trainer.target_steps - 1)

        pyplot.close()

        print("----------- FINISHED -----------")

        # Return last save path
        return full_path0





from . import common
import torch.nn as nn
import copy
import math

class ConvLikeBaseGen(common.LayerGen):
    def computeSize(self, layer, Hin, Win, kernel_size, padding, stride, dilation):
        Hout = (Hin + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1) // stride[0] + 1
        Wout = (Win + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1) // stride[1] + 1

        return Hout, Wout

    def kernel_size_helper(self, kernel_size, input_size, dilation = None):
        out_kernel_size = []
        for i in range(len(kernel_size)):
            if kernel_size[i] == -1:
                ks = input_size[i + 1]
                if dilation != None:
                    ks = int(math.ceil(ks / dilation[i]))

                out_kernel_size += [ks]
            else:
                out_kernel_size += [kernel_size[i]]
        out_kernel_size = tuple(out_kernel_size)
        return out_kernel_size

class ConvLikeGen(ConvLikeBaseGen):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, *args, **kwargs):
        super(ConvLikeGen, self).__init__(*args, **kwargs)
        if isinstance(kernel_size, int):
            kernel_size = [kernel_size, kernel_size]
        if isinstance(dilation, int):
            dilation = [dilation, dilation]
        self.conv_kwargs = {"in_channels" : in_channels,
                            "out_channels" : out_channels,
                            "kernel_size" : kernel_size,
                            "stride" : stride,
                            "padding": padding,
                            "dilation" : dilation,
                            "groups" : groups,
                            "bias" : bias}

class Conv2dGen(ConvLikeGen):
    def build(self, input_size):
        self.conv_kwargs["kernel_size"] = self.kernel_size_helper(self.conv_kwargs["kernel_size"], input_size, self.conv_kwargs["dilation"])
        layer = nn.Conv2d(*self.args, **self.conv_kwargs)
        ret = self.build_helper(layer, input_size)
#        print(input_size, layer, ret[0], ret[1])
        return ret

class ConvTranspose2dGen(ConvLikeGen):
    def set_linked_conv(self, conv):
        self.__linked_conv__ = conv

    def computeSize(self, layer, Hin, Win, kernel_size, padding, stride, dilation):
        Hout = (Hin - 1) * stride[0] - 2 * padding[0] + kernel_size[0] + layer.output_padding[0]
        Wout = (Win - 1) * stride[1] - 2 * padding[1] + kernel_size[1] + layer.output_padding[1]
        return Hout, Wout

    def build(self, input_size):
        if hasattr(self, "__linked_conv__"):
            self.conv_kwargs["kernel_size"] = self.__linked_conv__.conv_kwargs["kernel_size"]
        self.conv_kwargs["kernel_size"] = self.kernel_size_helper(self.conv_kwargs["kernel_size"], input_size, self.conv_kwargs["dilation"])
        layer = nn.ConvTranspose2d(*self.args, **self.conv_kwargs)
        return self.build_helper(layer, input_size)

class ConvTransposeHelper(object):
    def __init__(self, conv_array, decoder_input_channels):
        self.conv_array = conv_array
        self.decoder_input_channels = decoder_input_channels

    def run(self):
        out_array = []

        for i, c in enumerate(self.conv_array):
            if isinstance(c, tuple):
                name = c[0]
                c = c[1]
            else:
                name = None
            if (isinstance(c, Conv2dGen)):
                last_conv_index = i

        for i, c in enumerate(self.conv_array):
            if isinstance(c, tuple):
                name = c[0]
                c = c[1]
            else:
                name = None

            if (isinstance(c, Conv2dGen)):
                kwargs = copy.deepcopy(c.conv_kwargs)
                kwargs["out_channels"] = c.conv_kwargs["in_channels"]
                if i != last_conv_index:
                    in_channels = c.conv_kwargs["out_channels"]
                else:
                    in_channels = self.decoder_input_channels
                kwargs["in_channels"] = in_channels
#                print(i, last_conv_index, in_channels, kwargs)

                ct = ConvTranspose2dGen(**kwargs)
                ct.set_linked_conv(c)
                insert_index = 0
            else:
                insert_index = 1
                ct = copy.deepcopy(c)

            if name != None:
                ct = (name + "_tr", ct)

            out_array.insert(insert_index, ct)

        return out_array



import torch.nn as nn
from . import common

class BatchNorm2dGen(common.LayerGen):
    def build(self, input_size):
        layer = nn.BatchNorm2d(input_size[0])
        return layer, input_size


class BatchNorm1dGen(common.LayerGen):
    def build(self, input_size):
        layer = nn.BatchNorm1d(input_size[0])
        return layer, input_size



class UpsampleGen(LayerGen):
    def build(self, input_size):
        layer = nn.Upsample(input_size)
        if layer.size is not None:
            output_size = layer.size
        else:
            output_size = []
            for inp in input_size:
                output_size += [int(inp * layer.scale_factor)]
        output_size = tuple(output_size)

        return layer, output_size



import collections
from . import common, io, combine
import torch.nn as nn
    
CHILD_PARENT = {}

def get_layer_child_parent(child):
    return CHILD_PARENT[id(child)]

def set_layer_child_parent(child, parent):
    CHILD_PARENT[id(child)] = parent
    
class ContainerLayer(object):
    def get_module_and_output(self, output_name):
        output_name_split = output_name.split(".")
        
        child_name = output_name_split[0]

        if len(child_name) == 0:
            return get_layer_child_parent(self).get_module_and_output(".".join(output_name_split[1:]))
        
        if child_name not in self._modules:
            knowns = ", ".join(self._modules.keys())
            raise Exception("Unknown name %s in object %s, know names are %s" % (child_name, self.__class__.__name__ , knowns))
        m = self._modules[child_name]

        if len(output_name_split) == 1:
            outputs = None
            if hasattr(self, "outputs"):
                outputs = self.outputs[output_name]
            return m, outputs
        else:
            return m.get_module_and_output(".".join(output_name_split[1:]))

    def set_output(self, output_name, batch):
        self.outputs[output_name] = batch

    def get_output(self, output_name = None):
        if output_name == None:
            return self.outputs[output_name]
        return self.get_module_and_output(output_name)[1]

    def get_module(self, output_name):
        return self.get_module_and_output(output_name)[0]

    def display(self, xp, timestamp, current_path = []):
        if xp.display_disable() or not xp.accept_display(timestamp):
            return
        for child_name, child in self._modules.items():
            if isinstance(child, ContainerLayer):
                child.display(xp, timestamp, current_path + [child_name])
                continue

            module_full_name = ".".join(current_path + [child_name])
            output = self.get_output(child_name)
            image = output.cpu().data.clone().numpy()

            xp.add_display(timestamp, "net." + module_full_name + "#out", image)

            if hasattr(child, "weight") and child.weight is not None:
                weight = child.weight
                if weight.is_sparse:
                    weight = weight.to_dense()
                image = weight.cpu().data.clone().numpy()
                xp.add_display(timestamp, "net." + module_full_name + "#w", image)

            if hasattr(child, "bias") and child.bias is not None:
                image = child.bias.data.clone().numpy()
                xp.add_display(timestamp, "net." + module_full_name + "#b", image)




class ContainerLayerGen(common.LayerGen):
    def __init__(self, *args, **kwargs):
        super(ContainerLayerGen, self).__init__(*args, **kwargs)
        self.child_and_output_sizes = {}
        self.prepared = False
        self.parent_assigned = False
        
    def build(self, *args, **kwargs):
#        print("build", self.__class__.__name__, id(self), self.prepared)
        if not self.prepared:
            self.prepare()
            self.prepared = True
        return self.build_(*args, **kwargs)

    def prepare(self):
        self.assign_parent()

    def enumerate_python_args(self):
        return enumerate(self.args)
        
    def enumerate_children(self):
        for i, c in self.enumerate_python_args():
            if isinstance(c, tuple):
                name = c[0]
                obj = c[1]
            else:
                name = "#%d" % i
                obj = c
            yield name, obj

    def find_child_name(self, child):
        for name, c in self.enumerate_children():
            if c == child:
                return name

    def assign_parent(self):
        if self.parent_assigned:
            return
        self.parent_assigned = True

        for name, obj in self.enumerate_children():
#            print("assign_parent", self.__class__.__name__, id(self), obj.__class__.__name__, id(obj))
            obj.parent = self
            if hasattr(obj, "assign_parent"):
                obj.assign_parent()

    def get_output_size(self, name):
        name_split = name.split(".")

        if len(name_split[0]) == 0:
            return self.parent.get_output_size(".".join(name_split[1:]))

        if name_split[0] not in self.child_and_output_sizes:
            raise Exception("Could not find %s in %s: name set is %s" % (name_split[0], self, [ k for k in self.child_and_output_sizes.keys()]))
                            
        child, output_size = self.child_and_output_sizes[name_split[0]]
        if len(name_split) == 1:
            return output_size
        else:
            return child.get_output_size(".".join(name_split[1:]))

    def add_child_and_output_size(self, name, child, output_size):
        self.child_and_output_sizes[name] = (child, output_size)


class Broadcast(nn.Module, ContainerLayer):
    def __init__(self, children, *args, **kwargs):
        super(Broadcast, self).__init__(*args, **kwargs)
        for c in children:
            self.add_module(c[0], c[1])

    def forward(self, *args):
        self.outputs = {}
        outputs = []
        for n in self._modules.items():
            module_name = n[0]
            module = n[1]
            output = module(*args)
            self.set_output(module_name, output)
            outputs += [output]

        return outputs[0]

class BroadcastGen(ContainerLayerGen):
    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            self.child_enumerator_object = kwargs
            # kwargs is used to list child names
            kwargs = {}
        else:
            self.child_enumerator_object = args
            
        super(BroadcastGen, self).__init__(*args, **kwargs)

    def enumerate_python_args(self):
        if isinstance(self.child_enumerator_object, dict):
            return enumerate(self.child_enumerator_object.items())
        else:
            assert(isinstance(self.child_enumerator_object, list) or isinstance(self.child_enumerator_object, tuple))
            return enumerate(self.child_enumerator_object)
    
    def input_size_flatten(self, input_size):
        i = 1
        for j in input_size:
            i *= j
        return i

    def build_(self, input_size = None):
        modules = []
        first_final_size = None
        
        for name, obj in self.enumerate_children():
            module, final_size = obj.build(input_size)
            if first_final_size == None:
                first_final_size = final_size

            self.add_child_and_output_size(name, obj, final_size)

            modules += [(name, module)]

        layer = Broadcast(modules, **self.kwargs)

        for name, m in modules:
            set_layer_child_parent(m, layer)

        return layer, first_final_size

class Sequential(nn.Sequential, ContainerLayer):
    def forward(self, *args):
        self.outputs = {}

        batch = None
        for index, n in enumerate(self._modules.items()):
            module_name = n[0]
            module = n[1]

            if isinstance(module, combine.MultiInputModule):
#                print('Multi input')
                input_names = module.get_input_names()
                inputs = []
                input_shapes = []
                for input_name in input_names:
#                    print(module_name, module)
                    inpt = self.get_output(input_name)
                    inputs += [inpt]
                    input_shapes += [inpt.shape]
                
#                print("Sequential forward", module_name, input_shapes)
                batch = module(*inputs)
            else:
#                print("Sequential forward", module_name, batch.shape if batch is not None else None)

#                print('Not multi', module.__class__)
                if index == 0:
                    batch = module(*args)
                else:
                    batch = module(batch)

            self.set_output(module_name, batch)
        self.set_output(None, batch)
        return batch

class SequentialGen(ContainerLayerGen):    
    def build_(self, input_size = None):
        od = collections.OrderedDict()

        seq = Sequential()

        for name, obj in self.enumerate_children():
            if name in seq._modules:
                raise Exception("Duplicated module name %s" % name)
            input_names = obj.get_input_names()
            if input_names != None:
                input_size = []
                for input_name in input_names:
#                    print("SequentialGen", name, input_name)
                    input_size += [self.get_output_size(input_name)]

            last_input_size = tuple(input_size) if input_size is not None else None
            layer, input_size = obj.build(input_size)
#            print(name, last_input_size, input_size)

            self.add_child_and_output_size(name, obj, input_size)
            if(name in od):
                raise Exception("Layer %s is already in sequence" % name)
            seq._modules[name] = layer
            set_layer_child_parent(layer, seq)
            
        return seq, input_size




import cv2
import numpy

class World(object):
    def __init__(self, xp, resolution):
        super(World, self).__init__()
        self.image = numpy.zeros((resolution[0], resolution[1], 3), numpy.uint8)
        self.root = WorldObject(self, None)
        self.root.parent = self
        self.resolution = resolution
        self.xp = xp

    def size(self):
        return self.resolution

    def center(self):
        return (self.resolution[0] // 2, self.resolution[1] // 2)

    def add_child(self, o):
        self.root.add_child(o)

    def clear(self):
        self.image.fill(0)

    def update(self, timestamp):
        self.clear()
#        print("world draw", timestamp)
        self.root.draw(self.image)
        self.xp.add_display(timestamp, "world", self.image)
        self.root.run(timestamp)
        self.root.debug_draw(self.image)
        self.xp.add_display(timestamp, "world.debug", self.image)

    def get_image(self):
        return self.image

    def save_snapshot(self, file_name):
        cv2.imwrite(file_name, self.image)

    def position(self):
        return numpy.array([0,0])

    def add_display(self, timestamp, name, image):
        self.xp.add_display(timestamp, name, image)

    def get_world_path(self):
        return []

class WorldObject(object):
    def __init__(self, world, translation = None, name = None):
#        super(WorldObject, self).__init__()
        self.world = world
        if translation == None:
            self.translation = numpy.array([0,0])
        else:
            self.translation = numpy.array(translation)
        self.children = []
        self.name = name

    def get_world_path(self):
        return self.parent.get_world_path() + [self.name]

    def add_display(self, timestamp, name, image):
        path = [f for f in self.get_world_path() + [name] if f != None]
        path = ".".join(path)
        self.world.add_display(timestamp, path, image)

    def position(self):
        return self.translation + self.parent.position()

    def add_child(self, c):
        c.parent = self
        self.children += [c]

    def run(self, timestamp):
        for c in self.children:
            c.run(timestamp)

    def draw_self(self, image):
        pass

    def draw(self, image):
        self.draw_self(image)
        for c in self.children:
            c.draw(image)

    def debug_draw_self(self, image):
        pass

    def debug_draw(self, image):
        self.debug_draw_self(image)
        for c in self.children:
            c.debug_draw(image)

class Rect(WorldObject):
    def __init__(self, world, translation, size, color):
        super(Rect, self).__init__(world, translation)
        self.size = size
        self.color = color

    def draw(self, image):
        # Filled rectangle
        pos1 = tuple(self.position())
        pos2 = (pos1[0] + self.size[0], pos1[1] + self.size[1])
        cv2.rectangle(image, pos1, pos2, self.color, -1, cv2.LINE_4)


class Circle(WorldObject):
    def __init__(self, world, translation, radius, color):
        super(Circle, self).__init__(world, translation)
        self.radius = radius
        self.color = color

    def draw(self, image):
        # Filled rectangle
        pos1 = tuple(self.position())
        cv2.circle(image, pos1, self.radius, self.color, -1, cv2.LINE_4)













import os
from . import xp, brain
import shutil

class Cryo(object):
    def __init__(self, path = None):
        """path is the place where all the best networks are saved"""
        if path == None:
            path = os.path.join(xp.Experiment.aura_xp_data_path(), "cryo")

        self.path = os.path.abspath(path)

    def load_network(self, name, *args, **kwargs):
        path = os.path.join(self.path, name)
        return brain.Net.network_create_and_load(path, *args, **kwargs)

    def save_last_network(self, xp_name, name, force = False):
        src = xp.Experiment.get_latest_path_for_name(xp_name)

        try:
            print("date/hour     : %s" % os.path.split(os.path.split(src)[0])[-1])
            print("max_epoch (k) : %d" % (int(os.path.split(src)[-1].split("_")[-1]) // 1000))
        except:
            pass

        dest = os.path.join(self.path, name)
        if os.path.exists(dest) and force:
            shutil.rmtree(dest)

        shutil.copytree(src, dest)
        shutil.copy(os.path.join(src, "../loss.txt"), os.path.join(dest, "loss.txt"))

class Utils(object):
    @staticmethod
    def inspect(filename, path_to_layer, layer_component):
        assert(path_to_layer.startswith("net."))
        path_to_layer = path_to_layer[4:]
        net = brain.Net.network_create_and_load(filename)
        layer = net.network.get_module(path_to_layer)
        component = getattr(layer, layer_component)

        print(component)






from . import organism, xp, ann, device
import numpy
import torch
import torch.nn as nn
from torch.autograd import Variable
import math
import os
import copy
import json
import shutil

class Hook():
    def __init__(self, module, backward=False):
        if backward==False:
            self.hook = module.register_forward_hook(self.hook_fn)
        else:
            self.hook = module.register_backward_hook(self.hook_fn)
        self.input = None
        self.output = None
        
    def hook_fn(self, module, input, output):
        self.input = input
        self.output = output
        
    def close(self):
        self.hook.remove()

class Net(nn.Module):
    state_sizes = []
    # input_sizes contains the size of all the inputs
    def __init__(self, xp, input_sizes, bare = False, *args, **kwargs):
        super(Net, self).__init__()
        self.xp = xp

        self.args = args
        self.kwargs = kwargs

        # Make sure input and state sizes have the right type
        self.input_sizes = self.size_sanitize(input_sizes)
        self.state_sizes = self.size_sanitize(self.get_state_sizes())

        # Create inputs/states holders : X input will be accessible through Inputs.X , and X state will be accessible through State.X
        self.inputs_gen = self.holder_create(self.input_sizes)
        self.states_gen = self.holder_create(self.state_sizes)
        
        # Build the network object, passing it the input_sizes and the generator for inputs
        self.network, output_size = self.get_network()

#        print(self.network, output_size)
        
        self.get_states()
        self.add_backward_hook()
        
        assert(isinstance(self.network, ann.Sequential))
        
        device.prepare_object(self.network)
        if not bare:
            self.prepare()

        if hasattr(self, "output_size"):
            if(output_size != self.output_size):
                raise Exception("Invalid size %s compared to declared one %s " % (output_size, self.output_size))

        self.output_size = output_size
        self.output_reset()
        self.state_reset()

    def enumerate_modules_rec(self, root, prefix = "", terminal_only = True):
        is_container = isinstance(root, ann.ContainerLayer)
        
        if not is_container or not terminal_only:
            if prefix != "":
                yield prefix
                
        if not is_container:
            return
        
        for module_name, module in root._modules.items():
            if prefix != "":
                new_prefix = prefix + "." + module_name
            else:
                new_prefix = module_name

            for r in self.enumerate_modules_rec(module, prefix = new_prefix, terminal_only = terminal_only):
                yield r
            
    def get_network(self):
        # Build the top level sequence, prefixed by state definition and input definition
        seq_gen_children = []
        if len(self.state_sizes) != 0:
            seq_gen_children += [("State", self.states_gen)]
        seq_gen_children += [("Inputs", self.inputs_gen)]
        seq_gen_children += self.get_sequence()                
        seq_gen = ann.SequentialGen(*seq_gen_children)

        return seq_gen.build()

    def get_states(self):
        self.states = []
        for n in self.enumerate_modules_rec(self.network, terminal_only = False):
            if n.split(".")[-1] == "State":
                self.states += [n]

    def add_backward_hook(self):
        self.hooks = {}
        for name in self.enumerate_modules_rec(self.network, terminal_only = True):
            module = self.network.get_module(name)
            self.hooks[name] = Hook(module,backward=True)

    def get_state_sizes(self):
        return self.state_sizes
        
    def size_sanitize(self, sizes):
        s = []
        if isinstance(sizes, list) or isinstance(sizes, tuple):
            enumerator = sizes
        elif isinstance(sizes, dict):
            enumerator = sizes.items()
        else:
            raise Exception("Unknown type for sizes:", sizes)
        for name, size in enumerator:
            if not isinstance(size, int):
                assert(isinstance(size, list) or isinstance(size, tuple))
                size = [int(s) for s in size]
            else:
                size = (size,)
            s += [(name, size)]
#        print("size_sanitize output", s)
        return s
    
    def holder_create(self, sizes):
        holders_gen = [(name, ann.TensorHolderGen(size)) for name, size in sizes]
        return ann.BroadcastGen(*holders_gen)        

    def set_input(self, name, value):
        constant_holder = self.network._modules["Inputs"].get_module(name)
        constant_holder.set(value)

    def get_input(self, name):
        constant_holder = self.network._modules["Inputs"].get_module(name)
        return constant_holder.get()
        
    def display(self, xp, timestamp):
        for name, m in self._modules.items():
            if isinstance(m, ann.ContainerLayer):
                m.display(xp, timestamp)
        for module_name, hook in self.hooks.items():
            if hook.input is None:
                continue
            
            for t_index, t in enumerate(hook.input):
                if t is not None:
                    display_name = "net." + module_name + "#b%s" % t_index
#                    print(display_name, t_index, t.shape)
                    xp.add_display(timestamp, display_name, t.cpu().data.clone().numpy())
#            print(module_name, hook.input)

    def get_output(self):
        if not hasattr(self, "net_output"):
            self.output_reset()

        return self.net_output

    def output_reset(self):
        batch_output_size = (1,) + self.output_size
        self.net_output = device.zeros(batch_output_size)

    def state_reset(self):
        for state_name in self.states:
            holder = self.network.get_module(state_name)
        
            for tensor_holder_name, tensor_holder in holder._modules.items():
                tensor_holder.reset()                
                
    def forward(self):
        self.net_output = self.network()
        return self.net_output

    def prepare(self):
        # Do some final work, like adding some layers, for example for recurring networks when used standalone,
        # and not in combination with other networks
        pass

    def save(self, path):
        # TEMPORARY
        return        
        net_path = os.path.join(path, "network")
        with open(net_path, "wb") as f :
            torch.save(self, f)

        statedict_path = os.path.join(path, "statedict")
        with open(statedict_path, "wb") as f:
            torch.save(self.state_dict(), f)

        meta_path = os.path.join(path, "meta")
        xp_name = self.xp.get_xp_module()
        meta_dict = {"input_sizes":list(self.input_sizes),
                     "output_size":list(self.output_size), "xp_module":xp_name}
        if self.state_sizes != None:
            meta_dict["state_sizes"] = self.state_sizes
        
        with open(meta_path, "w") as f:
            f.write(json.dumps(meta_dict))

    def load(self, path = None):
        # If no path is given, load the latest network trained for this experiment
        if path == None:
            path = self.xp.get_latest_path()
        print("Net loading from %s" % path)
        # From https://pytorch.org/tutorials/beginner/saving_loading_models.html#saving-loading-model-for-inference
        with open(os.path.join(path, "statedict"), "rb") as f:
            state_dict =  torch.load(f)
        self.load_state_dict(state_dict, strict = True)

    @staticmethod
    def network_create(path, *args, **kwargs):
        meta_path = os.path.join(path, "meta")
        with open(meta_path, "r") as f:
            meta = json.loads(f.read())
        xp_module_name = meta["xp_module"]

        import importlib
        module = importlib.import_module(xp_module_name)
        # This is needed when using live coding stuff, otherwise the module is cached and the module.Net() constructor triggers an error
        importlib.reload(module)
        inputs_size = meta["inputs_size"]
        net = module.Net(None, inputs_size, *args, **kwargs)
        if "states_size" in meta:
            net.states_size = meta["states_size"]        

        return net

    @staticmethod
    def network_output_size(path, *args, **kwargs):
        net = Net.network_create(path, *args, **kwargs)
        return net.output_size

    @staticmethod
    def network_create_and_load(path, *args, **kwargs):
        net = Net.network_create(path, *args, **kwargs)
        net.load(path)

        return net

class Brain(organism.Organ):
    def __init__(self, world, translation, net):
        super(Brain, self).__init__(world, translation, "brain")
        self.net = net
        
    def run(self, timestamp):
        self.output = device.tensor2numpy(self.net.forward()).flatten()
        
    def set_input(self, name, value):
        value = device.tensor(value)
        self.net.set_input(name, value)

    def get_input(self, name):
        return self.net.get_input(name)
        
    def get_output(self, name):
        if name == "hand":
            ret = numpy.array(self.output)[:2]
            return ret
        elif name == "eye":
            ret = numpy.array(self.output)[:2]
            return ret
        elif name == "eye_confidence":
            return None
            out = numpy.array(self.output)
            if out.shape[0] > 2:
                ret = math.sqrt(math.exp(out[2]))
            else:
                ret = None
            return ret

    def save(self, path):
        self.net.save(path)

    def reset(self):
        self.net.output_reset()
        self.net.state_reset()




from . import world, cache, ann
import numpy
import cv2
import math

class Organ(world.WorldObject):
    def __init__(self, world, translation, name):
        super(Organ, self).__init__(world, translation, name)


# Old defaults
#    eye_reach = 32
#    retina_scale_down = 2
#    retina_final_scale_down = 2
#    scale_mode = 0
#    centered_eye = False
#    noise_level = 0.03


class EyeConfig(object):
    def __init__(self, reach = 0.5, retina_scale_down = 1, retina_final_scale_down = 16, scale_mode = 2, centered = False, noise_level = 0.03):
        self.reach = reach
        self.retina_scale_down = retina_scale_down
        self.retina_final_scale_down = retina_final_scale_down
        self.scale_mode = scale_mode
        self.centered = centered
        self.noise_level = noise_level
        

class Eye(Organ):
    def __init__(self, world, translation, config):               
        super(Eye, self).__init__(world, translation, "eye")
        self.color = (128, 128, 255)

        # Base resolution is world size
        self.input_resolution = world.size()
        self.scale_mode = config.scale_mode
        self.noise_level = config.noise_level
        if config.centered:
            multiplier = 1
        else:
            multiplier = 2
        self.tmp_retina_resolution = [multiplier * self.input_resolution[0], multiplier * self.input_resolution[1]]
        assert(self.tmp_retina_resolution[0] % 2 == 0)
        assert(self.tmp_retina_resolution[1] % 2 == 0)

        # Scale in two phases : first while doing retina fovea remapping, second simple scaling down with averaging of pixels
        self.retina_scale_down = config.retina_scale_down
        self.retina_final_scale_down = config.retina_final_scale_down
        # Temporary retina
        self.tmp_retina = numpy.zeros((self.tmp_retina_resolution[0], self.tmp_retina_resolution[1], 3), numpy.float)

        resolution = numpy.array(self.tmp_retina_resolution, numpy.int)
        resolution //= self.retina_scale_down * self.retina_final_scale_down
        self.resolution = (3, resolution[0], resolution[1])

        # Build the static remapping used for fovea
        self.build_remap()
        self.confidence_radius = None
        
    def set_output_result(self, o, function_name, function_args):
        self.output_result_callback = [o, function_name, function_args]

    def call_output_result(self, image):
        info = self.output_result_callback
        fun = getattr(info[0], info[1])
        fun(image, *info[2])
        
    def clear(self):
        """Clear the image where the world is pasted into (with translation)"""
        self.tmp_retina.fill(0)

    @staticmethod
    @cache.disk_cache(reset = False)
    def build_remap_impl(shape, scale_mode, retina_scale_down):
        center = numpy.array(shape) / 2.0
        retina_mapx = numpy.zeros(shape, numpy.float32)
        retina_mapy = numpy.zeros(shape, numpy.float32)
        scale = math.pow(shape[0] / 2.0, - scale_mode) * math.pow(retina_scale_down, scale_mode + 1)
        for i in range(shape[0]):
            for j in range(shape[1]):
                pos = numpy.array([i, j], numpy.float)
                pos -= center
                pos *= scale * math.pow(cv2.norm(pos), scale_mode)
                pos += center
                retina_mapy[i,j] = pos[0]
                retina_mapx[i,j] = pos[1]
        return retina_mapx, retina_mapy
        
    def build_remap(self):
        """Build the cv2.remap map"""
        shape = (self.tmp_retina.shape[0], self.tmp_retina.shape[1])


        self.retina_mapx, self.retina_mapy = self.build_remap_impl(shape, self.scale_mode, self.retina_scale_down)

    def run(self, timestamp):
        """Run the eye acquisition, putting the result in self.retina"""
        self.clear()
        image = self.world.get_image()

        image = numpy.array(image, numpy.float) / 255.0

        #Crop the world image into self.tmp_retina
        position = self.position()
        left = max(0, position[0] - self.tmp_retina.shape[0] // 2)
        top = max(0, position[1] - self.tmp_retina.shape[1] // 2)
        right = min(image.shape[0], position[0] + self.tmp_retina.shape[0] // 2)
        bottom = min(image.shape[1], position[1] + self.tmp_retina.shape[1] // 2)

        crop = image[top:bottom, left:right]

        xpos = left - (position[0]  - self.tmp_retina.shape[0] // 2)
        ypos = top - (position[1]  - self.tmp_retina.shape[1] // 2)

        self.tmp_retina[ypos:ypos + bottom - top, xpos:xpos + right - left] = crop

        # Remap self.tmp_retina into a new image
        pre_retina = cv2.remap(self.tmp_retina, self.retina_mapx, self.retina_mapy, interpolation = cv2.INTER_AREA)

        # Crop the center of the retina
        shape = pre_retina.shape

        r = self.retina_scale_down
        if r != 1:
            e = shape[0] // r // 2
            rmin = (r - 1) * e
            rmax = (r + 1) * e
            pre_retina = pre_retina[rmin:rmax, rmin:rmax]

        # Resize the retina to its final size
        if self.retina_final_scale_down == 1:
            self.retina = pre_retina
        else:
            final_scale = 1.0 / self.retina_final_scale_down
            self.retina = cv2.resize(pre_retina, None, fx = final_scale, fy = final_scale, interpolation = cv2.INTER_AREA)

        self.retina = numpy.array(self.retina, numpy.float)
        r = numpy.random.random_sample(self.retina.shape) * self.noise_level
        self.retina += r
        self.retina *= (1.0/ (1.0 + self.noise_level))

        self.add_display(timestamp, "retina", self.retina)

        image = numpy.transpose(self.retina, (2, 0, 1))
        shape = image.shape
        image = image.reshape([-1, shape[0], shape[1], shape[2]])
        image = numpy.array(image, dtype=numpy.float32)

        self.call_output_result(image)

    def debug_draw_self(self, image):
        pos = tuple(self.position())
#        print("debug_draw_self", pos)
        pos1 = (pos[0] - 2, pos[1] - 2)
        pos2 = (pos[0] + 2, pos[1] + 2)
        cv2.rectangle(image, pos1, pos2, self.color, -1, cv2.LINE_4)
        if self.confidence_radius != None:
            cv2.circle(image, pos, int(self.confidence_radius), (0,255,0))

    def draw_self(self, image):
        pos = tuple(self.position())
        pos1 = (pos[0] - 2, pos[1] - 2)
        pos2 = (pos[0] + 2, pos[1] + 2)
#        cv2.rectangle(image, pos1, pos2, (255, 0, 255), -1, cv2.LINE_4)

class Hand(Organ):
    def __init__(self, world, translation):
        self.color = (0,255,255)
        super(Hand, self).__init__(world, translation, "hand")

    def debug_draw_self(self, image):
        pos = tuple(self.position())
        pos1 = (pos[0] - 3, pos[1] - 3)
        pos2 = (pos[0] + 3 , pos[1] + 3)
        cv2.rectangle(image, pos1, pos2, self.color, -1, cv2.LINE_4)


class Organism(world.WorldObject):
    def __init__(self, world, translation):
        super(Organism, self).__init__(world, translation, "org")

class DummyBrain(Organ):
    def __init__(self, world, organism):
        super(DummyBrain, self).__init__(world, None)
        self.organism = organism
        self.hand_position = numpy.array([0.0,0.0])
        self.eye_position = numpy.array([0.0,0.0])

    def run(self, timestamp):
        random = numpy.random.random(size=(2,)) * 0.05 - 0.025
        self.hand_position += random
        random = numpy.random.random(size=(2,)) * 0.02 - 0.01
        self.eye_position += random

    def get_output(self, name):
        if name == "hand":
            return self.hand_position
        elif name == "eye":
            return self.eye_position        

class HandConfig(object):
    def __init__(self, reach = 0.5):
        self.reach = reach
    
class Alpha(Organism):
    has_hand = True
    has_eye = True
    
    def __init__(self, world, translation = None, configs = {}, *args, **kwargs):        
        world_size = world.resolution
        if translation is None:
            translation = list(numpy.array(world.resolution) // 2)
            
        super(Alpha, self).__init__(world, translation)
        self.configs = configs
        
        if self.has_hand:
            hand_config = self.configs.get("hand", HandConfig())
            self.hand_reach = self.build_reach(world_size, hand_config.reach)

            self.hand = Hand(world, None)            
            self.add_child(self.hand)
                
        if self.has_eye:
            eye_config = self.configs.get("eye", EyeConfig())
            self.eye_reach = self.build_reach(world_size, eye_config.reach)
        
            self.eye = Eye(world, None, eye_config)
            self.eye.set_output_result(self, "set_brain_input", ["image"])        
            self.add_child(self.eye)
            

    def build_reach(self, world_size, reach):
        return int(world_size[0] * reach)

    def set_brain_input(self, data, name):
        # Data should be a numpy image
        self.brain.set_input(name, data)
        
    def save(self, path):
        print("Saving to", path)
        self.brain.save(path)

    def set_brain(self, brain):
        self.brain = brain
        self.add_child(self.brain)

    def normalize_output(self, value, scale, bounds):
        t = numpy.multiply(value, scale)
        t = numpy.ndarray.astype(t, numpy.int)
        t = numpy.minimum(t, bounds)
        t = numpy.maximum(t, - bounds)
        return t

    def reset(self):
#        print("organism reset")
        if self.has_hand:
            self.hand.translation = numpy.array([0,0])
        if self.has_eye:
            self.eye.translation = numpy.array([0,0])
        self.brain.reset()

    def run(self, timestamp):
        super(Alpha, self).run(timestamp)
        if self.has_eye:
            hand = self.brain.get_output("hand")
            self.hand.translation = self.normalize_output(hand, self.hand_reach, self.hand_reach)

        if self.has_eye:
            eye = self.brain.get_output("eye")
            self.eye.translation = self.normalize_output(eye, self.eye_reach, self.eye_reach)

        eye_confidence = self.brain.get_output("eye_confidence")
        if eye_confidence is not None:
            self.eye.confidence_radius = eye_confidence * self.eye_reach[0]

    def create_brain(self):
        return DummyBrain(world, self)








import torch
import torch.nn as nn
from . import common, conv, utils
from .. import device


class MultiInputModule(nn.Module):
    def __init__(self, input_names, *args, **kwargs):
        self.input_names = input_names
        super(MultiInputModule, self).__init__(*args, **kwargs)

    def get_input_names(self):
        return self.input_names


class MultiInputModuleGen(common.LayerGen):
    def __init__(self, input_names, *args, **kwargs):
        if not isinstance(input_names, list) and not isinstance(input_names, tuple):
            input_names = [input_names] + list(args)
            args = []
        self.input_names = input_names
        super(MultiInputModuleGen, self).__init__(*args, **kwargs)

    def get_input_names(self):
        return self.input_names

class Get(MultiInputModule):
    def __init__(self, input_names, *args, **kwargs):
        super(Get, self).__init__(input_names, *args, **kwargs)
        
    def forward(self, *args):
        ret = args[0]
        return ret
    
class GetGen(MultiInputModuleGen):
    def __init__(self, input_name, *args, **kwargs):
        super(GetGen, self).__init__([input_name], *args, **kwargs)
        
    def build(self, input_sizes):
        layer = Get(self.input_names, *self.args, **self.kwargs)
        size = input_sizes[0]
            
        return layer, size

class Concat(MultiInputModule):
    def forward(self, *args):
        inputs = []
        for y in args:
            inputs += [y.view(1, -1)]

        if len(inputs) == 1:
            output = inputs[0]
        else:
            output = torch.cat(inputs, 1)
        return output
    
class ConcatGen(MultiInputModuleGen):
    def input_size_flatten(self, input_size):
        i = 1
        for j in input_size:
            i *= j
        return i

    def build(self, input_sizes):
        layer = Concat(self.input_names, *self.args, **self.kwargs)

        size = 0
        for input_size in input_sizes:
            size += self.input_size_flatten(input_size)

        return layer, (size,)

class TwoWayConv2d(MultiInputModule):
    def __init__(self, input_names, conv_layer):
        super(TwoWayConv2d, self).__init__(input_names)
        self.conv_layer = conv_layer
        
    def forward(self, *args):
        c = self.conv_layer
        shape = c.weight.shape
        weight = args[1].reshape(shape)
        
        return torch.nn.functional.conv2d(args[0], weight, bias = None, stride=c.stride, padding=c.padding, dilation=c.dilation, groups=c.groups)
            
class TwoWayConv2dGen(MultiInputModuleGen, conv.ConvLikeGen):
    def __init__(self, input_names, *args, **kwargs):
        self.input_names = input_names
        conv.ConvLikeGen.__init__(self, *args, **kwargs)
        
    def build(self, input_sizes):
        self.conv_kwargs["kernel_size"] = self.kernel_size_helper(self.conv_kwargs["kernel_size"], input_sizes[0], self.conv_kwargs["dilation"])

        conv_layer = nn.Conv2d(*self.args, **self.conv_kwargs)
        conv_layer, output_size = self.build_helper(conv_layer, input_sizes[0])
        
        layer = TwoWayConv2d(self.input_names, conv_layer)
        
        return layer, output_size            


class MaskedOperation(MultiInputModule):
    def __init__(self, input_names, masks = None, *args, **kwargs):
        if masks is not None:
            assert(len(masks) == len(input_names))            
            self.masks = [device.tensor(m) for m in masks]
        else:
            self.masks = None
        super(MaskedOperation, self).__init__(input_names, *args, **kwargs)

class MaskedOperationGen(MultiInputModuleGen):
    def __init__(self, input_names, masks = None, *args, **kwargs):
        if masks is not None:
            assert(len(masks) == len(input_names))
            self.masks = [device.tensor(m) for m in masks]
        else:
            self.masks = None
        super(MaskedOperationGen, self).__init__(input_names, *args, **kwargs)
    
class Add(MaskedOperation):
    def forward(self, *args):        
        output = args[0]
        if self.masks is not None:
            output = output.mul(self.masks[0])
            
        for i, y in enumerate(args[1:]):
            if self.masks is not None:
                y = y.mul(self.masks[1 + i])

            output = output.add(y)

        return output

class AddGen(MaskedOperationGen):
    def build(self, input_sizes):
        layer = Add(self.input_names, self.masks, *self.args, **self.kwargs)

        input_size0 = tuple(input_sizes[0])

        for input_size in input_sizes[1:]:
            if(input_size0 != tuple(input_size)):
                print("AddGen incorrect input sizes !!", input_size0, input_size)
            assert(input_size0 == tuple(input_size))

        return layer, input_size0

class Mul(MultiInputModule):
    def forward(self, *args):
        output = args[0]
        for i, y in enumerate(args[1:]):
            output = output.mul(y)

        return output

class MulGen(MultiInputModuleGen):
    def build(self, input_sizes):
        layer = Mul(self.input_names, *self.args, **self.kwargs)

        input_size0 = tuple(input_sizes[0])

        for input_size in input_sizes[1:]:
            if(input_size0 != tuple(input_size)):
                print("MulGen incorrect input sizes !!", input_size0, input_size)
            assert(input_size0 == tuple(input_size))

        return layer, input_size0

    

class Cross(MultiInputModule):
    def forward(self, *args):
        shape0 = args[0].shape
        shape1 = args[1].shape

        s0 = (shape0[0], utils.product(shape0[1:]), 1)
        s1 = (shape1[0], 1, utils.product(shape1[1:]))

        args0 = args[0].reshape(s0)
        args1 = args[1].reshape(s1)
        output = args0 * args1

        output = output.reshape(shape0[0], s0[1] * s1[2])
        
        return output

class CrossGen(MultiInputModuleGen):
    def build(self, input_sizes):
        assert(len(input_sizes) == 2)
        layer = Cross(self.input_names, *self.args, **self.kwargs)

        m = 1
        for input_size in input_sizes:
            m *= utils.product(input_size)

        return layer, (m,)



class LogEntry(object):
    def add(self, index, value):
        raise NotImplementedError("Implement this in subclasses")

    def process(self):
        pass

class Log(object):
    entry_factory = LogEntry

    def __init__(self, name):
        self.name = name
        self.entries = []

    def new_entry(self):
        return self.entry_factory()

    def append_new_entry(self, entry):
        self.entries += [entry.process()]

    def get_entries(self):
        return self.entries

class Recorder(object):
    def __init__(self, xp):
        self.logs = {}
        self.log_entries = None
        self.xp = xp

    def add_log(self, log):
        assert(log.name not in self.logs)
        self.logs[log.name] = log

    def start_new_log_entry(self):
#        print("start_new_log_entry")
        if self.log_entries != None:
            self.process_log_entries()
        self.log_entries = {}
        for log_name, log in self.logs.items():
            self.log_entries[log_name] = log.new_entry()

    def process_log_entries(self):
        for log_name, log_entry in self.log_entries.items():
            self.logs[log_name].append_new_entry(log_entry)

    def add(self, log_name, index, value):
        self.log_entries[log_name].add(index, value)

    def display(self):
        for log_name, log in self.logs.items():
#            print(log_name, log.get_entries())
            self.xp.add_display(-1, "recorder." + log_name, log.get_entries())







from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()

setup(name='auran',
      version='0.1.1',
      description='auran is a python package built on top of pytorch to ease creation of artificial neural networks.',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.0',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      keywords='',
      url='',
      author='Francois Lagunas',
      author_email='francois.lagunas@m4x.org',
      license='MIT',
      packages=['auran', 'auran.ann'],
      install_requires=['click'],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      entry_points={
          'console_scripts': ['auran_run=auran.command_line:main'],
      },
      include_package_data=True,
      zip_safe=False)



import collections
import numpy
import matplotlib.ticker
import matplotlib.pyplot as pyplot
import math
import _thread
import time


class Figure(object):
    def __init__(self, config, image_name, image, *args, **kwargs):
        self.config = config
        self.name = image_name
        self.image = image
        if not isinstance(self.image, numpy.ndarray):
            self.image = numpy.array(self.image)
        self.args = args
        self.kwargs = kwargs
        self.size = numpy.array(image).shape

    def run(self):
        self.plot_call(self.image, *args, **kwargs)

    def draw(self, sp):
        #                sp.set_xticks([i for i in range(figure.image.shape[0])])
#                sp.set_yticks([i for i in range(figure.image.shape[0])]

        sp.tick_params(axis='both', which='major', labelsize=4, length=2, pad=1, reset = False)

        sp.set_title(self.name, size=10)

class ImageFigure(Figure):
    def draw(self, sp):
        super(ImageFigure, self).draw(sp)
        scale = 1

        for shape_axis in self.image.shape:
            while shape_axis > scale * 10:
                scale *= 10

        majorLocator = matplotlib.ticker.MultipleLocator(scale)
        minorLocator = matplotlib.ticker.MultipleLocator(1)

        sp.xaxis.set_major_locator(majorLocator)
#        sp.xaxis.set_minor_locator(minorLocator)
        sp.yaxis.set_major_locator(majorLocator)
#        sp.yaxis.set_minor_locator(minorLocator)

        pyplot.imshow(self.image, interpolation = "none")

class PlotFigure(Figure):
    def draw(self, sp):
        super(PlotFigure, self).draw(sp)
        non_one = 0
        for s in self.size:
            if s != 1:
                non_one += 1

        legend = []
        image = self.image

        transpose = self.config.transpose
        if transpose != None and transpose != False:
            if transpose is True:
                transpose = [1, 0]
            image = numpy.transpose(image, transpose)

        if non_one <= 1:
            image = image.flatten()
            legend += [self.config.legend_prefix]
        else:
            for i in range(self.size[1]):
                legend += [self.config.legend_prefix + str(i)]

        if self.config.legend != None:
            legend = self.config.legend

        pyplot.plot(image)

        if len(legend) > 1:
            sp.legend(legend, loc='upper left')

class ColorBarFigure(Figure):
    def draw(self, sp):
        try:
            pyplot.colorbar(cax=sp, orientation = "vertical")
        except Exception as e :
            print(e)

        frame_index = 10
        pyplot.text(0.0, -0.15, "frame=%d" % frame_index, fontsize=12)

class Config(object):
    figure_factory = None
    # Is the object bound to a story time ? (True, except for LossPlotConfig that does not depend on specific story time)
    frame_based = True

    def image_names(self, image_name):
        return [image_name]

    def new_figures_from_image(self, image_name, image):
        ret = []
        for image_name, image in self.process_image(image_name, image):
            ret += [self.figure_factory(self, image_name, image)]
        return ret

    def process_image(self, image_name, image):
        return [(image_name, image)]

class PlotConfig(Config):
    figure_factory = PlotFigure

    def __init__(self, legend = None, legend_prefix = "graph", transpose = None):
        super(PlotConfig, self).__init__()
        self.legend_prefix = legend_prefix
        self.transpose = transpose
        self.legend = legend

class LossPlotConfig(PlotConfig):
    figure_factory = PlotFigure
    frame_based = False

    def image_names(self, image_name):
        return [(image_name + "#" + sn) for sn in self.sub_names]

    def __init__(self, sub_names = ["loss"], legend_prefix = "loss"):
        super(LossPlotConfig, self).__init__()
        self.legend_prefix = legend_prefix
        self.sub_names = sub_names
        self.transpose = None

    def process_image(self, image_name, image):
        initial_loss_image = numpy.array(image)
        initial_name = image_name

        loss_images = []
        if len(initial_loss_image.shape) == 3:
            last_dim_size = initial_loss_image.shape[2]
            if last_dim_size > len(self.sub_names):
                raise Exception("Not enough names in LossPlotConfig : loss last dimension is %d, and sub_names are %s" % (last_dim_size, self.sub_names))
            for i in range(last_dim_size):
                subimage_name = initial_name  + "#" + self.sub_names[i]
                subimage = initial_loss_image[:,:,i]
                loss_images += [(subimage_name, subimage)]
        else:
            if len(initial_loss_image.shape) == 1 and initial_loss_image.shape[0] != 0:
                assert(len(self.sub_names) == 1)

            return [(image_name + "#" + sn, initial_loss_image) for sn in self.sub_names]

        return loss_images

class ColorBarConfig(Config):
    figure_factory = ColorBarFigure
    frame_based = False

class ImageConfig(Config):
    figure_factory = ImageFigure

class DefaultImageConfig(ImageConfig):
    pass

class RGBImageConfig(ImageConfig):
    def process_image(self, image_name, image):
        image = numpy.transpose(image[0], (1, 2, 0))
        return [(image_name, image)]

class FlattenImageConfig(ImageConfig):
    def __init__(self, transpose = False, *args, **kwargs):
        self.transpose = transpose
        super(FlattenImageConfig, self).__init__(*args, **kwargs)

    def process_image(self, image_name, image):
        shape = image.shape
        if len(shape) == 4:
            if shape[3] == 1 :
                new_shape = [shape[0] * shape[1], shape[2]]
            else:
                new_shape = [shape[0] * shape[1] * shape[2], shape[3]]
        if len(shape) == 3:
            new_shape = [shape[0] * shape[1], shape[2]]
        elif len(shape) == 2:
            new_shape = shape
        elif len(shape) == 1:
            new_shape = [shape[0], 1]

        image = image.reshape(new_shape)

        if self.transpose:
            image = image.transpose()
        
        return [(image_name, image)]

class SquareImageConfig(ImageConfig):
    def __init__(self, divider, parts = None, transpose = False):
        self.divider = divider
        self.parts = parts or [1]
        self.transpose = transpose

    def image_name_build(self, image_name, index):
        return image_name + "@%d" % index

    def image_names(self, image_name):
        ret = []
        for i in range(len(self.parts)):
            ret += [self.image_name_build(image_name, i)]
        return ret

    def process_image(self, image_name, image):
        # input should be displayed as a square, with info[1] channels
        divider = self.divider
        parts = self.parts

        if len(image.shape) == 4:
            s = image.shape
            image = image.reshape([s[0] * s[1], s[2] * s[3]])
        if self.transpose:
            image = image.transpose()
#        print("process_image", image_name, image.shape)
        size0 = image.shape[1]

        size0 //= sum(parts)

        start = 0
        images = []
        for part_index, p in enumerate(parts):
            size = size0 * p
            size = int(math.sqrt(size))
            amount = size * size
            stop  = start + amount
            img = image[:,start:start + amount]
            start = stop
            img = img.reshape([divider * size, size])
            images += [(self.image_name_build(image_name, part_index), img)]
        if (stop != image.shape[1]):
            if not(hasattr(self, "debug_first")):
                print("STOPPING BEFORE IMAGE END", stop, image.shape, image_name)
        return images

def pgcd(a,b):
    """pgcd(a,b): calcul du 'Plus Grand Commun Diviseur' entre les 2 nombres entiers a et b"""
    while b != 0:
        r=a%b
        a,b=b,r
    return a

def ppcm(a,b):
    """ppcm(a,b): calcul du 'Plus Petit Commun Multiple' entre 2 nombres entiers a et b"""
    if (a==0) or (b==0):
        return 0
    else:
        return (a*b)//pgcd(a,b)

class DashboardLayoutConfig(object):
    def __init__(self):
        self.whitelist = {}
        self.rows = collections.OrderedDict()
        self.rows_config = collections.OrderedDict()

    def add_display(self, row_name, name, config):
        if row_name not in self.rows:
            self.rows[row_name] = collections.OrderedDict()

        for n in config.image_names(name):
            self.rows[row_name][n] = config

        self.whitelist[name] = row_name

    def set_row_config(self, row_name, key, value):
        if row_name not in self.rows_config:
            self.rows_config[row_name] = {}
        self.rows_config[row_name][key] = value

    def get_row_config(self, row_name, key, default):
        return self.rows_config.get(row_name, {}).get(key, default)

    def get(self, name):
        row_name = self.whitelist.get(name)
        if row_name is None:
            return None

        ret = self.rows.get(row_name).get(name, None)
        if ret:
            return ret

        # Try subnames, as some images generate several figures, return the first one
        for figure_name, config in self.rows.get(row_name).items():
            if figure_name.startswith(name):
                return config

    def ppcm_cols(self):
        m = 1
        for name, c in self.rows.items():
            m = ppcm(m, len(c))
        return m

    def rows_total_size(self):
        total_size = 0
        for name, c in self.rows.items():
            total_size += self.get_row_config(name, "width", 1)
        return total_size

class DashboardFrame(object):
    def __init__(self, frame_index):
        self.frame_index = frame_index
        self.figures = {}

    def add_figure(self, figure):
        self.figures[figure.name] = figure

    def get_figure(self, name):
        return self.figures.get(name, None)

class DashboardFrameSet(object):
    def __init__(self, timestamp):
        self.timestamp = timestamp
        self.frames = []
        self.append(DashboardFrame(0))

    def append(self, frame):
        self.frames.append(frame)

    def add_display(self, timestamp, figure):
        if timestamp == -1:
            frame_index = 0
        else:
            frame_index = timestamp - self.timestamp

        while len(self.frames) <= frame_index:
            self.append(DashboardFrame(len(self.frames)))

        self.frames[frame_index].add_figure(figure)

    def get_frame_count(self):
        return len(self.frames)

    def get_frame(self, frame_index):
        return self.frames[frame_index]

    def get_figure_count(self):
        fcount = 0
        for f in self.frames:
            fcount = max(fcount, len(f.get_figures()))
        return fcount

class Dashboard(object):
    def __init__(self, config, backend):
        self.config = config
        self.backend = backend
        self.debug = False
        self.started = False
        self.started_timestamp = None

    def start_sequence(self, timestamp):
        if not self.started:
            self.new_frame_set = DashboardFrameSet(timestamp)
            self.started = True
            self.started_timestamp = timestamp

    def accept_display(self, timestamp):
        return self.started

    def display_figure_enabled(self, name):
        dashboard_config = self.config.get("dashboard/layout")
        if dashboard_config == None:
            return None
        return dashboard_config.get(name)

    def add_display(self, timestamp, image_name, image):
        if not self.started:
            return

        image = image.copy()
        figure_config = self.display_figure_enabled(image_name)
        if self.debug:
            print(image_name, figure_config)
        if figure_config is not None:
            figures = figure_config.new_figures_from_image(image_name, image)
            for figure in figures:
                if self.debug:
                    print(figure.name, timestamp)

                self.new_frame_set.add_display(timestamp, figure)

    def end_sequence(self):
        if not self.started:
            return

        frame_set = self.new_frame_set
        self.new_frame_set = None

        dashboard_config = self.config.get("dashboard/layout")
        self.backend.display(frame_set, dashboard_config)

        self.started = False

    def display_refresh(self, filename = None, frame_index = None):
        self.backend.display_refresh(filename, frame_index = frame_index)

class DashboardBackend(object):
    def display(self, frame_set):
         # To be implemented by subclasses
        pass

class MPLBackend(DashboardBackend):
    def __init__(self, config):
        self.mp_fig = None
        self.config = config
        self.last_time = 0.0
        self.frame_index = 0
        self.frame_set = None
        self.first = True

    def display(self, frame_set, dashboard_config):
        self.frame_set = frame_set
        self.dashboard_config = dashboard_config

        self.frame_index = 0
        self.display_refresh()

    def display_refresh(self, filename = None, frame_index = None):
        if self.frame_set == None:
            return

        if frame_index == None:
            t = time.time()
            if (t - self.last_time) < self.config.get("dashboard/story_refresh", 1.0):
                return
            self.last_time = t

            if self.frame_index >= self.frame_set.get_frame_count():
                self.frame_index = 0
                # Pause a bit : we should not return, but it creates a single frame pause
                return
            use_frame_index = self.frame_index
        else:
            use_frame_index = frame_index

        self.display_init(self.dashboard_config)

        self.display_core(self.frame_set, self.dashboard_config, use_frame_index)
        #        self.mp_fig.tight_layout()
        self.display_finish(self.dashboard_config, filename)

        if frame_index == None:
            self.frame_index += 1


    def display_init(self, dashboard_config):
        if self.mp_fig == None:
            figsize = self.config.get("dashboard/size", (12, 6))
            self.mp_fig = pyplot.figure(num=self.config.get("xp/name"), figsize=figsize)

        pyplot.clf()

    def display_core(self, frame_set, dashboard_config, frame_index):
        ppcm_cols = dashboard_config.ppcm_cols()
        rows_total_size = dashboard_config.rows_total_size()

        row_offset = 0
        sps = []
        for row_index, (row_name, row) in enumerate(dashboard_config.rows.items()):
            rowspan = dashboard_config.get_row_config(row_name, "height", 1)
            colspan = ppcm_cols // len(row.items())

            for col_index, (figure_name, figure_config) in enumerate(row.items()):
                sp = pyplot.subplot2grid((rows_total_size, ppcm_cols), (row_offset, col_index * colspan),
                                         rowspan=rowspan, colspan=colspan)
                sps += [sp]
                if figure_config.frame_based:
                    figure_frame_index = frame_index
                else:
                    figure_frame_index = 0
                figure = frame_set.get_frame(figure_frame_index).get_figure(figure_name)
                if figure is None:
                    valid_names = ",".join(frame_set.get_frame(figure_frame_index).figures.keys())
                    print("Invalid figure name %s, figure_frame_index = %d, valid names are: %s" % (figure_name, figure_frame_index, valid_names))
                    continue

                figure.draw(sp)
                try:
                    pyplot.colorbar()
                except:
                    pass

            row_offset += rowspan

    def pause(self, interval):
        # From https://stackoverflow.com/questions/45729092/make-interactive-matplotlib-window-not-pop-to-front-on-each-update-windows-7/45734500#45734500
        backend = pyplot.rcParams['backend']
        if backend in matplotlib.rcsetup.interactive_bk:
            figManager = matplotlib._pylab_helpers.Gcf.get_active()
            if figManager is not None:
                canvas = figManager.canvas
                if canvas.figure.stale:
                    canvas.draw()
                canvas.start_event_loop(interval)
                return
            
    def display_finish(self, dashboard_config, filename):
        margin = 0.05
        pyplot.subplots_adjust(left=margin, right=1.0-margin, top=1.0 - margin, bottom=margin, hspace = 0.2, wspace = 0.4)

        if self.first:
            pyplot.show(block=False)
            self.first = False

        if filename != None:
            self.pause(0.5)
            self.display_save(filename)
        else:
            self.pause(0.00000001)

    def display_save(self, filename):
        pyplot.savefig(filename)







import numpy
from torch.autograd import Variable
import torch.optim as optim
import torch.nn as nn
import torch
from . import recorder
import os
import os.path
import time
import math

class BaseStory(object):
    loss_info_parts = 1
    def __init__(self, trainer, duration, post_roll = 0):
        self.trainer = trainer
        self.start_time = None
        self.duration = duration + post_roll
        self.post_roll = post_roll
        self.loss_info = []

    def finished(self, time):
        if self.start_time == None:
            return False
        return (time - self.start_time) >= self.duration

    def story_init(self):
        """Called at time == 0"""
        self.trainer.reset()

    def story_update(self, story_time):
        raise NotImplementedError()

    def update(self, time):
        assert(not self.finished(time))
        if self.start_time == None:
            self.start_time = time
            self.story_init()

        reltime = time - self.start_time
        if reltime < self.duration - self.post_roll:
            self.story_update(reltime)

    def get_reference(self, story_time):
        # Should return a tensor
        raise NotImplementedError()

    def get_debug_loss(self, story_time):
        return math.sqrt(self.total_loss)

    def post_update(self, time):
        reltime = time - self.start_time

        if reltime < self.duration - self.post_roll:
            self.story_post_update(reltime)

        if reltime < self.duration - self.post_roll - 1:
            self.story_step_finish(reltime)
        elif reltime == self.duration - self.post_roll - 1:
            self.story_finish(reltime)

    def story_post_update(self, story_time):
        self.reference = self.get_reference(story_time)

        # Display

        if not self.trainer.xp.display_disable():
            numpy_reference = self.reference.cpu().data.numpy().flatten()
            numpy_output = self.trainer.network.get_output().cpu().data.numpy().flatten()
            self.trainer.xp.add_display(story_time + self.start_time, "story.reference", numpy_reference)
            self.trainer.xp.add_display(story_time + self.start_time, "story.output", numpy_output)
            if numpy_output.size != numpy_reference.size:
                raise Exception("Output %s and reference %s should have the same size" % (numpy_output.size, numpy_reference.size))
            
            output_and_reference = numpy.stack([numpy_output, numpy_reference], axis = 1)
                
            self.trainer.xp.add_display(story_time + self.start_time, "story.out_and_ref", output_and_reference)
        
        self.total_loss = self.trainer.compute_loss(story_time + self.start_time, self.reference)

        assert(len(self.loss_info) == story_time)

        self.loss_info.append(self.get_debug_loss(story_time))

    def story_step_finish(self, story_time):
        self.trainer.story_step_finish(story_time + self.start_time)

    def story_finish(self, story_time):
        self.trainer.story_finish(story_time + self.start_time)
        self.record_stats(story_time)

    def record_stats(self, story_time):
        pass

    def set_brain_input(self, name, value):
        # TODO : clean this mess
        return self.trainer.xp.organism.brain.set_input(name, value)


class Trainer(object):
    # TODO : override this in subclasses
    story_factory = None
    loss_scaling = 1.0
    loss_aggregate = 1000
    all_step_optimize = False

    def __init__(self, xp, target_steps = 1, post_roll = 0):
        self.xp = xp
        self.criterion = self.criterion_get()
        self.optimizer = None
        self.target_steps = target_steps
        self.post_roll = post_roll
        self.story = None
        self.loss_reset()
        self.loss_history = []
        self.recorder = recorder.Recorder(self.xp)
        self.recorder_init()
        self.recorder.start_new_log_entry()
        self.create_story(0)
        self.story_counter = 0
        self.run_start_time = time.time()

    def recorder_init(self):
        pass

    def create_story(self, timestamp):
        self.story = self.story_factory(self, self.target_steps, self.post_roll)
        self.xp.start_training_sample(timestamp)

    def reset(self):
        self.xp.organism.reset()
        self.network = self.xp.alpha.brain.net

    def loss_reset(self):
        self.loss_info = None

    def loss_fill(self, story):
        loss_info = numpy.array(story.loss_info)

        if self.loss_info is None:
            self.loss_info = numpy.zeros(loss_info.shape, dtype = numpy.float32)

        self.loss_info += loss_info

    def criterion_get(self):
        return nn.MSELoss(reduction='sum')

    def optimizer_init(self, parameters):
        #return optim.Adam(parameters, lr=0.01)
        #return optim.Adadelta(parameters, lr = 0.5)
        #return optim.ASGD(parameters, lr = 0.001)

        #return optim.SGD(parameters, lr=0.00001, momentum=0.5)

        return optim.RMSprop(parameters, lr = 0.0001)

    def optimizer_get(self):
        if self.optimizer == None:
            parameters = filter(lambda p: p.requires_grad, self.network.parameters())
            self.optimizer = self.optimizer_init(parameters)
            self.optimizer.zero_grad()

        return self.optimizer

    def compute_loss(self, time, reference):
        outputs = self.network.get_output()
        ref_outputs = Variable(reference)

        self.loss = self.criterion(outputs, ref_outputs)
        
        return self.loss.data.item()

#    def update_organism(self):
#        # TODO : remove this ??
#        self.xp.organism.update_eye()

    def update(self, t):
        if self.story.finished(t):
            self.story_counter += 1
            if self.story_counter % self.loss_aggregate == 0:
                avg = (self.loss_info / self.loss_aggregate) * self.loss_scaling
                self.loss_history += [avg]
                print("ELAPSED=", time.time() - self.run_start_time , "SPEED =", self.story_counter / (time.time() - self.run_start_time), "\nLOSS after story_count=", self.story_counter, "\n", avg)
                self.loss_reset()
                self.recorder.start_new_log_entry()

            self.xp.add_display(-1, "global.loss", self.loss_history)
            self.recorder.display()

            self.xp.finish_training_sample()

            self.loss_fill(self.story)

            self.create_story(t)

        self.story.update(t)

    def save(self, path):
        if not (self.loss_history is None):
            out = numpy.array(self.loss_history)
            if out.shape != (0,):
                out = out.reshape((out.shape[0], -1))

                indexes = numpy.array([ [i * self.loss_aggregate] for i in range(out.shape[0])])

                out = numpy.append(indexes, out, axis = 1)

                header = "shape = " + str(out.shape)
                loss_file_tmp = os.path.join(path, "loss.txt.tmp")
                loss_file = os.path.join(path, "loss.txt")

                numpy.savetxt(loss_file_tmp, out, header = header)
                os.rename(loss_file_tmp, loss_file)


    def post_update(self, t):
        self.story.post_update(t)

    def optimize(self, t, retain_graph = False):
        optimizer = self.optimizer_get()
        optimizer.zero_grad()

        if self.loss is not None:
            self.loss.backward(retain_graph = retain_graph)
            optimizer.step()


    def story_step_finish(self, t):
        if self.all_step_optimize:
            self.optimize(t, retain_graph = True)
        
        self.network.display(self.xp, t)

    def story_finish(self, t):
        self.optimize(t)
        self.network.display(self.xp, t)



import importlib.util
import os.path

"""Config keys:
  dashboard.update_freq : frequency in hz, default 1hz
"""

class Config(object):
    def __init__(self, config_path):
        """config_path is the path to a python file with a "fill_config" method that will be called to fill this object.
This allows some code to generate the config, and to reload it periodically after the user has changed it.
         """
        self.config_path = config_path
        self.config_path_time_stamp = None
        self.info = {}
        if config_path != None:
            self.reload()

    def regularize(self, path):
        if not isinstance(path, list):
            path = path.split("/")
        return path

    def get(self, path, default = None):
        path = self.regularize(path)
        d = self.info
        for p in path[:-1]:
            d = d.get(p, {})

        return d.get(path[-1], default)

    def set(self, path, value):
        path = self.regularize(path)
        d = self.info
        for i, p in enumerate(path[:-1]):
            if p not in d:
                d[p] = {}
            d = d[p]

        d[path[-1]] = value


    def dump(self, root = None):
        if root == None:
            root = self.info

        for k, v in self.info.items():
            if isinstance(v, dict):
                self.dump(v)
            else:
                print("%s -> %s" % (k,v))


    def reload(self):
        mtime = os.path.getmtime(self.config_path)
        if mtime == self.config_path_time_stamp:
#            print("No change")
            return
        self.config_path_time_stamp = mtime

        # Reload the module. From https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        spec = importlib.util.spec_from_file_location("__aura_config__", self.config_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        # Call the function fill_config on the module
        self.info = {}
        new_config = Config(None)
        module.fill_config(new_config)
        self.info = new_config.info
#        self.dump()



from .common import *
from .io import *
from .linear import *
from .conv import *
from .nonlinear import *
from .container import *
from .combine import *
from .norm import *
from .pool import *
from .utils import *
from .distributions import *
from .dropout import *



import click
from . import cryo_save_impl, inspect_impl

@click.group()
@click.pass_context
def cli(ctx):
    ctx.obj = {}

@cli.command()
@click.pass_context
@click.option('--force', '-f', is_flag = True, default = False)
@click.argument('xp_name')
@click.argument('cryo_name')
def cryo_save(ctx, xp_name, cryo_name, force):
    cryo_save_impl(xp_name, cryo_name, force)

@cli.command()
@click.pass_context
@click.argument('filename')
@click.argument('layer')
@click.argument('component')
def inspect(ctx, filename, layer, component):
    """FILENAME is a path to a directory where the statdict and meta files are stored
       LAYER is a path to some layer, for example net.Conv2d

       COMPONENT name of part of the layer, like weights or bias"""
    inspect_impl(filename, layer, component)

    
def main():
    return cli()



from . import common, combine, container
from .. import device
import torch.nn as nn

class ReshapeView(nn.Module):
    def __init__(self, shape):
        self.shape = shape
        super(ReshapeView, self).__init__()

    def forward(self, x):
        ret = x.view(1, *self.shape)
        return ret

class ReshapeViewGen(common.LayerGen):
    def __init__(self, shape):
        self.shape = shape

    def build(self, input_size):
        a = 1
        for s in input_size:
            a *= s

        b = 1
        for s in self.shape:
            b *= s

        if a != b:
            raise Exception("ReshapeViewGen: input %s number of elements %s is different from specified shape %s who have %s elements" % (input_size, a, self.shape, b))

        layer = ReshapeView(self.shape)

        return layer, self.shape


class WrapperGen(common.LayerGen):
    def __init__(self, net):
        from .. import brain
        assert(isinstance(net, brain.Net))
        self.net = net

    def build(self, input_size):
#        if input_size != self.net.input_size:
#            raise Exception("Invalid input_size %s : different from wrapped network %s" % (input_size, self.input_size))
        assert(isinstance(self.net.network, container.ContainerLayer))
        return self.net.network, self.net.output_size

class Extract(nn.Module):
    def __init__(self, start, end, *args, **kwargs):
        super(Extract, self).__init__(*args, **kwargs)
        self.start, self.end = start, end
        
    def forward(self, batch):
        return batch[:,self.start:self.end]

class ExtractGen(common.LayerGen):
    def __init__(self, start, end, *args, **kwargs):
        super(ExtractGen, self).__init__(*args, **kwargs)
        self.start, self.end = start, end
        
    def build(self, input_size):
        output_size = list(input_size)
        output_size[0] = self.end - self.start
        layer = Extract(self.start, self.end)
        return layer, tuple(output_size)

    
class TensorHolder(nn.Module):
    def __init__(self, size,  *args, **kwargs):
        super(TensorHolder, self).__init__()
        self.reset([1] + list(size))

    def set(self, value):
        self.value = value

    def get(self):
        return self.value

    def reset(self, size = None):
        if size is None:
            size = self.value.shape
        self.value = device.zeros(size)

    def forward(self, *args):
        return self.value

class TensorHolderGen(common.LayerGen):
    def __init__(self, size, *args, **kwargs):
        self.size = tuple(size)

    def build(self, input_size):
        layer = TensorHolder(self.size)
        return layer, self.size

class TensorUpdater(nn.Module):
    def __init__(self, name_map, dest_root_name = None, *args, **kwargs):
        # Name map is a list of couples input_name, holder_name
        self.name_map = name_map
        self.dest_root_name = dest_root_name
        assert(self.name_map is not None)

        super(TensorUpdater, self).__init__()

    def forward(self, batch):
        # Get parent for source, and possiblity for destination
        parent = container.get_layer_child_parent(self)
        if self.dest_root_name is not None:
            dest_root = parent.get_module(self.dest_root_name)
        else:
            dest_root = parent

        for src, dest in self.name_map:
            # src name is relative to parent
            src_output = parent.get_output(src)
            # dest name is relative to parent or dest_root if specified
            dest_module = dest_root.get_module(dest)
            old_value = dest_module.get()
            old_value_shape = old_value.shape
            dest_module.set(src_output.view(old_value_shape))
            
        return batch
        
class TensorUpdaterGen(common.LayerGen):
    def __init__(self, name_map, dest_root_name = None, **kwargs):
        # Name map is a list of couples input_name, holder_name
        self.name_map = name_map
        self.dest_root_name = dest_root_name
        super(TensorUpdaterGen, self).__init__()

    def build(self, input_size):
        if self.dest_root_name != None:
            # Get the parent path
            self_name = ".".join(self.full_name().split(".")[1:])
            # Substract the parent path from the dest full path, to get a relative path
            dest_root_name = self.dest_root_name[len(self_name):]
        else:
            dest_root_name = None
        layer = TensorUpdater(self.name_map, dest_root_name)

        return layer, input_size

class RelativeDestTensorUpdaterGen(TensorUpdaterGen):
    def __init__(self, networkWrapper, *args, **kwargs):
        self.networkWrapper = networkWrapper
        # holder_name will be setup in build
        super(RelativeDestTensorUpdaterGen, self).__init__(*args, **kwargs)

    def build(self, input_size):
        self.dest_root_name = self.networkWrapper.full_name()
        return super(RelativeDestTensorUpdaterGen, self).build(input_size)



import tempfile
import os.path
import functools
import pickle
import inspect
import contextlib


def get_class_that_defined_method(meth):
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
           if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects

class disk_cache(object):
    def __init__(self, reset = False):
        self.reset = reset
        
    def __call__(self, func):        
        @functools.wraps(func)
        def memoized_func(*args, **kwargs):
            tmp_dir = tempfile.gettempdir()
            aura_path = os.path.join(tmp_dir, "aura")
            os.makedirs(aura_path, exist_ok = True)

            key = "%s.%s-%s-%s" % (str(get_class_that_defined_method(func).__name__),
                                str(func.__name__),
                                str(args),
                                str(kwargs))
            full_path  = os.path.join(aura_path, key)

#            print("memoized", full_path)
            if not os.path.exists(full_path) or self.reset:
                o = func(*args, **kwargs)
                data = pickle.dumps(o)

                with contextlib.suppress(FileNotFoundError):
                    os.remove(full_path)
                    
                with tempfile.NamedTemporaryFile(dir=os.path.dirname(full_path)) as f:
#                    print("START write %s" % full_path)
                    f.write(data)
                    os.link(f.name, full_path)
#                    print("END write %s", full_path)


            with open(full_path, "rb") as f:
#                print("START load %s" % full_path)
                data = f.read()
                o = pickle.loads(data)
#                print("END load")

            return o

        return memoized_func

    




from . import common
from .. import device
import torch
import torch
import torch.nn as nn
import torch.nn.parameter as parameter
import random
        
class Scale(nn.Module):
    def __init__(self, scale, *args, **kwargs):
        super(Scale, self).__init__(*args, **kwargs)
        self.scale = scale
        
    def forward(self, batch):
        return batch * self.scale
    
class ScaleGen(common.LayerGen):
    def __init__(self, scale, *args, **kwargs):
        super(ScaleGen, self).__init__(*args, **kwargs)
        self.scale = scale
        
    def build(self, input_size):
        if input_size[0] != len(self.scale):
            raise Exception("Invalid length %s for scale parameter : should be %s." % (len(self.scale), input_size[0]))
        layer = Scale(device.prepare_object(device.tensor(self.scale)))
        return layer, input_size

class LogCascade(nn.Module):
    def __init__(self, precision, output_dimensions, *args, **kwargs):
        super(LogCascade, self).__init__(*args, **kwargs)
        self.precision = precision
        self.output_dimensions = output_dimensions
        self.weights = []
        w = 0.5
        for i in range(self.precision):
            self.weights += [w]
            w /= 2
        self.weights = device.prepare_object(device.tensor(self.weights))
        
    def forward(self, batch):
        batch = batch.reshape([-1, self.output_dimensions, self.precision])
        output = torch.matmul(batch, self.weights)
        
        return output
    
class LogCascadeGen(common.LayerGen):
    def __init__(self, output_dimensions, *args, **kwargs):
        super(LogCascadeGen, self).__init__(*args, **kwargs)
        self.output_dimensions = output_dimensions
        
    def build(self, input_size):
        p = product(input_size)
        if p % self.output_dimensions != 0:
            raise Exception("Invalid input size %s, product of input sizes must be a multiple of output_dimensions %s" % (p, self.output_dimensions))
        layer = LogCascade(p / output_dimensions, output_dimensions)
        return layer, input_size


class LinearBaseGen(common.LayerGen):
    def __init__(self, out_features, *args, **kwargs):
        self.out_features = out_features
        super(LinearBaseGen, self).__init__(*args, **kwargs)

    def build(self, input_size):
        in_features = input_size[-1]

        layer = self.Constructor(in_features, self.out_features, *self.args, **self.kwargs)

        output_size = tuple(input_size[:-1]) + (self.out_features,)

        if self.post_hook != None:
            self.post_hook(layer, output_size)

        return layer, output_size

class LinearGen(LinearBaseGen):
    Constructor = nn.Linear

class LinView(nn.Module):
    def __init__(self, truncate = 1):
        self.truncate = truncate
        super(LinView, self).__init__()

    def forward(self, x):
        size = x.size()
        t = [size[0]]

        for i in range(1, len(size) - self.truncate - 1):
            t += [size[i]]
        t += [-1]
        ret = x.view(*t)

        return ret

class LinViewGen(common.LayerGen):
    def __init__(self, truncate = 1):
        self.truncate = truncate

    def build(self, input_size):
        layer = LinView(self.truncate)
        size = input_size

        t = []
        for i in range(0, len(size) - self.truncate - 1):
            t += [size[i]]

        m = 1
        for i in range(max(0, len(size) - self.truncate - 1), len(size)):
            m *= size[i]

        t += [m]

        output_size = tuple(t)
        return layer, output_size


class OpParam(nn.Module):
    def __init__(self, size, collapse, distribute = True, normal_mean_init = 0.0, normal_std_init = 1.0, *args, **kwargs):
        super(OpParam, self).__init__()
        self.weight = nn.Parameter(torch.FloatTensor(size))
        # Is this the right way to initialize weights ?
        torch.nn.init.normal(self.weight, mean = normal_mean_init, std=normal_std_init)
        self.collapse = collapse
        self.distribute = distribute
        self.args = args
        self.kwargs = kwargs

    def forward(self, batch):
        weights_count = self.weight.shape[0]
        if self.collapse:
            batch_repeat = [1] * len(batch.shape)
            if self.distribute:
                batch_repeat[-1] *= weights_count
                weights_repeat = [s for s in batch.shape]
            else:
                assert(weights_count % batch.shape[-1] == 0)
                batch_repeat[-1] *= weights_count // batch.shape[-1]
                weights_repeat = [s for s in batch.shape]
                weights_repeat[-1] = 1
        else:
            batch_repeat = [1] * len(batch.shape) + [weights_count]
            if self.distribute:
                weights_repeat = [s for s in batch.shape] + [1]
            else:
                # TODO
                assert(False)
                weights_repeat = [s for s in batch.shape] + [1]

#        print(batch_repeat, batch.shape)
#        print(weights_repeat, self.weight.shape)
        batch = batch.repeat(*batch_repeat)
        weights = self.weight.repeat(*weights_repeat)
        
        return self.op(batch, weights)

    
class OpParamGen(common.LayerGen):
    def __init__(self, size, collapse = False, distribute = True, *args, **kwargs):
        self.size = size
        self.distribute = distribute
        self.collapse = collapse
        super(OpParamGen, self).__init__(*args, **kwargs)
        
    def build(self, input_size):
        layer = self.OpParam(self.size, self.collapse, self.distribute, *self.args, **self.kwargs)
        output_size = [s for s in input_size]
        if self.distribute:
            if self.collapse:        
                output_size[-1] *= self.size
            else:
                output_size += [self.size]
        else:
            output_size[-1] = self.size
        output_size = tuple(output_size)
        return layer, output_size


class AddParam(OpParam):
    def op(self, batch, weights):
        return batch + weights

class SubParam(OpParam):
    def op(self, batch, weights):
        return batch - weights
    
class MulParam(OpParam):
    def op(self, batch, weights):
        sign = self.kwargs.get("sign", 0)
        if sign== 1:
            weights = abs(weights)
        elif sign == -1:
            weights = -abs(weights)
            
        ret = batch * weights
        return ret

class SubParamGen(OpParamGen):
    OpParam = SubParam

class AddParamGen(OpParamGen):
    OpParam = AddParam
  
class MulParamGen(OpParamGen):
    OpParam = MulParam

class Neg(nn.Module):
    def forward(self, batch):
        return - batch

class Neg(common.LayerGen):
    Layer = Neg


class SparseLinear(nn.Module):
    def __init__(self, in_features, out_features = None, weight = True, bias = True, *args, **kwargs):
        super(SparseLinear, self).__init__()

        if not weight:
            assert(out_features == None)
            out_features = in_features

        self.in_features = in_features
        self.out_features = out_features
        
        if weight:
            self.weight = parameter.Parameter(self.create_weight(*args, **kwargs))
        else:
            self.register_parameter('weight', None)
            
        if bias:
            self.bias = parameter.Parameter(self.create_bias())
        else:
            self.register_parameter('bias', None)

    def create_bias(self):
        return device.zeros(self.out_features)        
            
    def create_sparse_weight_default(self, max_x, max_y, bands = 1, overlap = 0):
        # The caller guarantees that max_x < max_y
        assert(max_x <= max_y)
        ratio = (max_y / max_x) / bands

        x = []
        y = []
        for b in range(bands):
            for i in range(max_x):
                start = i * ratio
                end = (i + 1) * ratio

                for j in range(round(start - overlap / 2.0), round(end + overlap / 2.0)):
                    if j <= 0:
                        j += max_y

                    RAND = 4
                    # Wrap around, it's important so the first and last lines have two neighbours, and not just one
                    y_index = ((b * max_y) / bands + j + max_y + random.randint(-RAND, RAND)) % max_y 

                    x_index = i
                    
                    x += [x_index]
                    y += [y_index]

        return x, y 
                                
    def create_sparse_weight(self, *args, **kwargs):
        max_x = self.out_features
        max_y = self.in_features

        transpose = False
        if max_x > max_y:
            max_x, max_y = max_y, max_x
            transpose = True

        x, y = self.create_sparse_weight_default(max_x, max_y, *args, **kwargs)

        if transpose:
            return y, x
        else:
            return x, y
                              
        
    def create_weight(self, normal_mean_init = 0.0, normal_std_init = 1.0, *args, **kwargs):
        x, y = self.create_sparse_weight(*args, **kwargs)
        
        self.indexes = [x,y]
        self.values = torch.zeros(len(self.indexes[0]))

        # Is this the right way to initialize weight ?
        torch.nn.init.normal(self.values, mean = normal_mean_init, std=normal_std_init)

        self.values = torch.abs(self.values)

        self.indexes = torch.LongTensor(self.indexes)
        
        ret = torch.sparse.FloatTensor(torch.LongTensor(self.indexes),
                                       self.values,
                                       torch.Size([self.out_features, self.in_features]))

        ret = ret.coalesce()
        return ret
    
    def forward(self, input):
        if self.weight is not None:
            b = input[0].reshape(input[0].shape[0], -1)
            output = torch.sparse.mm(self.weight, b)
            output = output.reshape(output.shape[0])            
            
        if self.bias is not None:
            output += self.bias
            
        return output.reshape(1, -1)
        
class SparseLinearGen(LinearBaseGen):
    Constructor = SparseLinear
